#include <libunwind.h>
#include <signal.h>
#include <stdio.h>
#include <memory.h>

void backtrace(struct sigcontext *sc) 
{
	unw_cursor_t cursor; 
	unw_context_t uc;
	unw_word_t ip, sp;
	int err;
	fprintf(stderr, "Backtrace:\n");

	memcpy(&uc, sc, sizeof uc);
	if ((err = unw_init_local(&cursor, &uc)) != 0)
	{
		fprintf(stderr, "unw_init_local: ");
		switch (err) {
		case -UNW_EINVAL:
			fprintf(stderr, "UNW_EINVAL\n");
			break;
		case -UNW_EUNSPEC:
			fprintf(stderr, "UNW_EUNSPEC\n");
			break;
		case -UNW_EBADREG:
			fprintf(stderr, "UNW_EBADREG\n");
			break;
		default:
			fprintf(stderr, "Error %d\n", err);
			break;
		}
		return;
	}
	while ((err = unw_step(&cursor)) > 0) {
		unw_get_reg(&cursor, UNW_REG_IP, &ip);
		unw_get_reg(&cursor, UNW_REG_SP, &sp);
		fprintf(stderr, "ip = %lx, sp = %lx\n", (long) ip, (long) sp);
	}
	switch (err) {
	case 0:
		return;
#define ER(code) \
		case -code: fprintf(stderr, "unw_step: " #code "\n"); break
	ER(UNW_EUNSPEC);
	ER(UNW_ENOINFO);
	ER(UNW_EBADVERSION);
	ER(UNW_EINVALIDIP);
	ER(UNW_EBADFRAME);
	ER(UNW_ESTOPUNWIND);
	}
}
