#include <unistd.h>

#define likely(x)	__builtin_expect(!!(x), 1)
#define unlikely(x)	__builtin_expect(!!(x), 0)

#define DEBUG_EMU_BIT (1<<0)
#define DEBUG_EXCEPTION_BIT (1<<1)
#define DEBUG_SIGNAL_BIT (1<<2)
#define DEBUG_MM_BIT (1<<3)
#define DEBUG_IPC_BIT (1<<4)
#define DEBUG_SSC_BIT (1<<5)
#define DEBUG_STRACE_BIT (1<<6)
#define DEBUG_IO_BIT  (1<<7)
#define DEBUG_CTXT_BIT (1<<8)

#define abort() vm_panic(sc, "Abort at %s:%d\n", __FILE__, __LINE__)
//#define abort() kill(getpid(), SIGABRT)

//#define DEBUG (DEBUG_MM_BIT|DEBUG_SIGNAL_BIT|DEBUG_EXCEPTION_BIT)
//#define DEBUG (DEBUG_IPC_BIT)
#define DEBUG (0UL)

/*
 * Optimisation options
 */

#define RR_CHANGE_ALWAYS_PURGE  1 /* Old way: purge entire region on context switch */
#undef RR_CHANGE_ALWAYS_PURGE

#define OPTIMISE_SET_ADDRRANGE 1 /* Use ptrace(PTRACE_ADDRRANGE) to avoid syscalls during VMM operation */
//#undef OPTIMISE_SET_ADDRRANGE

#define OPTIMISE_PTRACE_SIGNALS 1 /* Don't ptrace signals */
//#undef OPTIMISE_PTRACE_SIGNALS

#define OPTIMISE_DISABLE_PTRACE 1 /* use PTRACE_VIRTUAL extension */
//#undef OPTIMISE_DISABLE_PTRACE

#define OPTIMISE_USE_MULTI_AS 1
//#undef OPTIMISE_USE_MULTI_AS

#if 1
# undef O_DIRECT
# define O_DIRECT 0
#endif

/*
 * To enable SYNC_IO needs changes in the guest too
 */
#define SYNC_IO			1
#undef SYNC_IO



static inline unsigned long read_itc(void)
{
	register unsigned long val;
	asm ("mov %0=ar.itc\n" :"=r"(val));
	return val;
}


/*
 * exception.c, signal.c
 */
#if DEBUG & DEBUG_EXCEPTION_BIT
# define DEBUG_EXCEPTION(fmt, args...) fprintf(stderr, "[%ld-%d] " fmt, read_itc(), getpid(), ## args)
#else
# define DEBUG_EXCEPTION(...) /* nothing */
#endif

/* emu.c */
#if DEBUG & DEBUG_EMU_BIT
# define DEBUG_EMU(fmt, args...) fprintf(stderr, "[%ld-%d] " fmt, read_itc(), getpid(), ## args)
#else
# define DEBUG_EMU(...)		/* nothing */
#endif

/*
 * mm.c
 */
#if DEBUG & DEBUG_MM_BIT
# define DEBUG_MM(fmt, args...) fprintf(stderr, "[%ld-%d] " fmt, read_itc(), getpid(), ## args)
#else
# define DEBUG_MM(...) /* nothing */
#endif

#if 0
# define DEBUG_DSM(args...) fprintf(stderr, args);
#else
# define DEBUG_DSM(...) /* nothing */
#endif

#if DEBUG & DEBUG_SSC_BIT
# define SSC_DEBUG 1
#else
# undef SSC_DEBUG
#endif


#if DEBUG & DEBUG_IPC_BIT
# define IPC_DEBUG(fmt, args...) fprintf(stderr, "[%ld-%d] " fmt, read_itc(), getpid(), ## args)
# define DEBUG_IPC
#else
# define IPC_DEBUG(fmt, args...)
# undef DEBUG_IPC
#endif

#if DEBUG & DEBUG_IO_BIT
# define IO_DEBUG(fmt, args...) fprintf(stderr, "[%ld-%d] " fmt, read_itc(), getpid(), ## args)
# define DEBUG_IO
#else
# define IO_DEBUG(fmt, args...)
# undef DEBUG_IO
#endif

#if DEBUG & DEBUG_CTXT_BIT
# define CTXT_DEBUG(fmt, args...)fprintf(stderr, "[%ld-%d] " fmt, read_itc(), getpid(), ## args)
# define DEBUG_CTXT
extern void dump_ctx(void);
#define DUMP_CTX() dump_ctx()
#define DUMP_MAPS() dump_maps()
#else
# define CTXT_DEBUG(fmt, args...) /**/
# undef DEBUG_CTXT
# define DUMP_MAPS() /**/
# define DUMP_CTX() /**/
#endif
