#define _LARGEFILE64_SOURCE
#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/signal.h>
#include "vm.h"
#include "debug.h"
#include "ssc.h"

#define KEYBOARD_INTR			3
#define NETWORK_INTR			8
#define SIMSCSI_INTR			9

extern char args[256];

extern unsigned long smp_rendezvous_ip, smp_rendezvous_gp;


long
do_ssc(struct sigcontext *sc, long arg0, long arg1, long arg2, long arg3, int nr)
{
#ifdef SYNC_IO
	struct disk_req *req;
	static struct disk_stat stat;
#endif
#ifdef DSM
	struct request request;
#endif
	char c;
	int r;

#ifdef SSC_DEBUG
	if (nr != SSC_PUTCHAR)
		fprintf(stderr, "[%d] SSC %d @ 0x%lx\n", cpu.id, nr, sc->sc_ip);
#endif

	switch (nr)
	{
		case SSC_GETCHAR:
			r = read(STDIN_FILENO, &c, 1);
			return (r > 0) ? c : 0;

		case SSC_PUTCHAR:
			c = (char)arg0;
			putchar(c);
			return 0;

		case SSC_OPEN:
#ifdef SSC_DEBUG
			fprintf(stderr, "SSC_OPEN %s\n", (char *)arg0);
#endif
			r = open((char *)arg0, O_RDWR|O_LARGEFILE|O_DIRECT|O_NOATIME);
#ifdef SSC_DEBUG
			if (r == -1)
				vm_perror("ssc_open %s", (char *)arg0);
#endif
			return r;

		case SSC_CLOSE:
			return close(arg0);

#if SYNC_IO
		case SSC_READ:
		{
			struct iovec iov[16];
			int i;
			int redo_direct = 0;

			req = (struct disk_req *)arg2;
#ifdef SSC_DEBUG
			fprintf(stderr, "SSC_READ, offset %lx, fd %d\n", arg3, (int)arg0);
#endif

			if (arg3 < 0)
				fprintf(stderr,"SSC_READ at offset %lx\n", arg3);
			stat.fd = arg0;
			stat.count = 0;
			if ((arg3 & 0x1ff) || (req[0].addr & 0x1ff) || req[0].len & 0x3ff) {
				fcntl(arg0, F_SETFL, O_NOATIME);
				redo_direct=1;

			}
			if (lseek(arg0, (off_t)arg3, SEEK_SET) != arg3)
				vm_perror("lseek");

			else while (arg1) {
				int max = arg1 > 16 ? 16 : arg1;
				int result;

				for (i = 0; i < max; i++) {
#ifdef SSC_DEBUG
					fprintf(stderr, "SSC_READ: at %p, size %lX\n",
						(void *)req[i].addr, (unsigned long)req[i].len);
#endif
					iov[i].iov_base = (void *)req[i].addr;
					iov[i].iov_len = req[i].len;
				}

				arg1 -= max;
				arg2 = (unsigned long)&req[max];

				while ((result = readv(arg0, iov, max)) == -1)
				{
#ifdef DSM
					if (errno != EFAULT)
					{
#endif
						vm_perror("read");
						break;
#ifdef DSM
					}
					mm_fetch(sc, req->addr & ~PAGE_MASK, 1, req->addr);
#endif
				}
				stat.count += result;
			}
			if (redo_direct) {
				fcntl(arg0, F_SETFL, O_DIRECT|O_NOATIME);
				redo_direct = 0;
			}
			return stat.count;
		}

		case SSC_WRITE:
		{
			struct iovec iov[16];
			int i;

			stat.fd = arg0;
			stat.count = 0;

			fcntl(arg0, F_SETFL, O_NOATIME);

			if (lseek(arg0, (off_t)arg3, SEEK_SET) != arg3)
				vm_perror("lseek");
			else 
			while (arg1) {
				int max = arg1 > 16 ? 16 : arg1;
				int result;


				req = (struct disk_req *)arg2;

				for (i = 0; i < max; i++) {
					iov[i].iov_base = (void *)req[i].addr;
					iov[i].iov_len = req[i].len;
				}

				arg1 -= max;
				arg2 = (unsigned long)&req[max];
				while ((result = writev(arg0, iov, max)) == -1)
				{
#ifdef DSM
					if (errno == EBADF)
					{
						printf("EBADF: sending request to %d\n", 1-cpu.id);
						request.type = IPC_SSC;
						request.subtype = nr;
						request.ssc.addr = req->addr;
						request.ssc.offset = arg3;
						request.ssc.fd = arg0;
						stat.count = request.ssc.len = req->len;
						ipc_call(sc, 1-cpu.id, &request);
						break;
					}
					else if (errno != EFAULT)
					{
#endif
						vm_perror("write");
						break;
#ifdef DSM
					}
					mm_fetch(sc, req->addr & ~PAGE_MASK, 0, req->addr);
#endif
				}
				stat.count += result;
			}
			fcntl(arg0, F_SETFL, O_NOATIME|O_DIRECT);
			return stat.count;
		}
		
		case SSC_WAIT_COMPLETION:
			/* currently I/O is synchronous */
			*(struct disk_stat *)arg0 = stat;
			return 0;
#else /* SYNC_IO */
		case SSC_READ:
		case SSC_WRITE:
			return queueio(arg0 & 0xffffffff,
			      arg0 >> 32,
			      arg1, arg3, nr == SSC_WRITE,
			      (struct disk_req *)arg2);

		case SSC_GET_COMPLETION:
			return aio_getcompletion();

		case SSC_WAIT_COMPLETION:
			return aio_wait_completion((struct disk_stat *)arg0);

		case SSC_NULL:
			IO_DEBUG("SSC_NULL: setting last_quiescent_req to %ld\n",
				 arg0);
			last_quiescent_req = arg0;
			deliver_interrupt(sc, simscsi_irq);
			return 0;
#endif /* SYNC_IO */
		case SSC_CONNECT_INTERRUPT:
			if (arg0 == KEYBOARD_INTR) {
				fcntl(STDIN_FILENO, F_SETOWN, getpid());
				fcntl(STDIN_FILENO, F_SETSIG, SIGKEYBD);
				fcntl(STDIN_FILENO, F_SETFL, fcntl(STDIN_FILENO, F_GETFL)|O_NONBLOCK|O_ASYNC);
#ifdef SSC_DEBUG

				fprintf(stderr, "Setting SIGIO for fd %d pid %d (console)\n",
					STDIN_FILENO, getpid());
#endif
				keyboard_irq = arg1;
			}
			else if (arg0 == NETWORK_INTR) {
				fcntl(network_fd, F_SETOWN, getpid());
				fcntl(network_fd, F_SETSIG, SIGNETWK);
				fcntl(network_fd, F_SETFL, fcntl(network_fd, F_GETFL)|O_NONBLOCK|O_ASYNC);
#ifdef SSC_DEBUG
				fprintf(stderr, "Setting SIGIO for fd %d pid %d (network)\n",
					network_fd, getpid());
#endif
				network_irq = arg1;
			}
			else if (arg0 == SIMSCSI_INTR)
				simscsi_irq = arg1;
			else fprintf(stderr, "Unknown irq %ld at IP=0x%lx\n",
				     arg0, sc->sc_ip);
#ifdef SSC_DEBUG
			fprintf(stderr, "Connect_irq: pid %d %ld->%d\n", getpid(), arg0, (int)arg1);
#endif
			return 0;

		case SSC_GET_ARGS:
			strcpy((char *)arg0, args);
			return strlen((char *)arg0);

		case SSC_GET_TOD:
			return gettimeofday((struct timeval *)arg0, NULL);

		case SSC_EXIT:
			vm_ssc_shutdown();

		case SSC_NETDEV_PROBE:
			memcpy((void *)arg1, network_ethaddr, sizeof(network_ethaddr));
			return network_fd;

		case SSC_NETDEV_SEND:
			write((int)arg0, (void *)arg1, arg2);
			return 0;

		case SSC_NETDEV_RECV:
			r = read((int)arg0, (void *)arg1, arg2);
			return (r > 0) ? r : 0;

		case SSC_SET_RENDEZVOUS:
			smp_rendezvous_ip = arg0;
			smp_rendezvous_gp = arg1;
			return 0;

		case SSC_CONSOLE_INIT:
		case SSC_LOAD_SYMBOLS:
		case SSC_NETDEV_ATTACH:
		case SSC_NETDEV_DETACH:
			return 0;

		case SSC_GET_MEMSIZE:
			*(unsigned long *)arg1 = get_memsize((unsigned long *)arg0);
			printf("GET_MEMSIZE: %lx->%lx\n",
			       *(unsigned long *)arg0,
			       MEM_SIZE);
			return 0;

		case SSC_PANIC:
			vm_panic(sc, "guest requested panic\n");
			exit(0);

		default:
			fprintf(stderr, "Unimplemented SSC %d\n", nr);
	}

	return -1;
}

unsigned long get_memsize(unsigned long * arg0)
{
  if (*arg0 < MEM_START)
    *arg0 = MEM_START;
  return (MEM_SIZE - *(unsigned long *)arg0) >> 12;
}
