#include <sys/types.h>
#include <setjmp.h>
//#define DSM
#define SMP
//#define NODES			2
extern int ncpu;
#define NODES ncpu
#define MAXCPU			128
#define EMULATOR_STAT		1
#include "vmlayout.h"
#define PIB_BASE		(0xfee00000)
#define VM_PORT_BASE		30000

#undef PAGE_SHIFT
#undef PAGE_MASK
#if CONFIG_4k
#define PAGE_SHIFT		(12)
#else  /* 16k pages */
#define PAGE_SHIFT		(14)
#endif

#define PAGE_MASK		((1 << PAGE_SHIFT) - 1)
#define NODE_SHIFT		(28)
#define NODE_MASK		((1 << NODE_SHIFT) - 1)

#define SWAP(a,b,temp)		{ temp = a; a = b; b = temp; }
#define BITS(val,from,n)	((val >> from) & ((1UL << n) - 1))

#define CR_DCR				0
#define CR_ITM				1
#define CR_IVA				2
#define CR_PTA				8
#define CR_IPSR				16
#define CR_ISR				17
#define CR_IIP				19
#define CR_IFA				20
#define CR_ITIR				21
#define CR_IIPA				22
#define CR_IFS				23
#define CR_IIM				24
#define CR_IHA				25
#define CR_LID				64
#define CR_IVR				65
#define CR_TPR				66
#define CR_EOI				67
#define CR_IRR0				68
#define CR_IRR1				69
#define CR_IRR2				70
#define CR_IRR3				71
#define CR_ITV				72
#define CR_PMV				73
#define CR_CMCV				74
#define CR_LRR0				80
#define CR_LRR1				81

#define PSR_MFL_BIT			 4
#define PSR_MFH_BIT			 5

#define PSR_IC_BIT			13
#define PSR_I_BIT			14
#define PSR_DFH_BIT			19
#define PSR_SI_BIT			23
#define PSR_DB_BIT			24
#define PSR_CPL_OFFSET			32
#define PSR_ID_BIT			37
#define PSR_DD_BIT			39
#define PSR_SS_BIT			40
#define PSR_RI_OFFSET			41
#define PSR_BN_BIT			44

#define PSR_IC				(1UL << PSR_IC_BIT)
#define PSR_I				(1UL << PSR_I_BIT)
#define PSR_DFH				(1UL << PSR_DFH_BIT)
#define PSR_SI				(1UL << PSR_SI_BIT)
#define PSR_DB				(1UL << PSR_DB_BIT)
#define PSR_CPL				(3UL << PSR_CPL_OFFSET)
#define PSR_ID				(1UL << PSR_ID_BIT)
#define PSR_DD				(1UL << PSR_DD_BIT)
#define PSR_SS				(1UL << PSR_SS_BIT)
#define PSR_RI				(3UL << PSR_RI_OFFSET)
#define PSR_BN				(1UL << PSR_BN_BIT)

#define PSR_MFL				(1UL << PSR_MFL_BIT)
#define PSR_MFH				(1UL << PSR_MFH_BIT)

#define TPR_MMI				(1UL << 16)

#define CFM_VALID			(1UL << 63)

/* Exceptions */
#define ALT_DTLB_MISS		0x1000
#define ALT_ITLB_MISS		0x0c00
#define BREAK_EXCEPTION		0x2c00
#define D_ACCESS_RIGHTS		0x5300
#define D_ACCESS_BIT		0x2800
#define D_KEY_MISS_EXCEPTION	0x1c00
#define NESTED_DTLB_MISS	0x1400
#define DTLB_MISS		0x0800
#define DEBUG_VECTOR		0x5900
#define DIRTY_BIT_VECTOR	0x2000
#define DISABLED_FPH		0x5500
#define EXT_IRQ_VECTOR		0x3000
#define FP_FAULT		0x5c00
#define FP_TRAP			0x5d00
#define GENERAL_EXCEPTION	0x5400
#define IA32_EXCEPTION		0x6900
#define IA32_INTERCEPT		0x6a00
#define IA32_IRQ		0x6b00
#define I_ACCESS_RIGHTS		0x5200
#define I_ACCESS_BIT		0x2400
#define I_KEY_MISS		0x1800
#define ITLB_MISS		0x0400
#define KEY_PERMS		0x5100
#define LPTT			0x5e00 /* Lower privilege transfer trap */
#define NAT_CONSUMPTION		0x5600
#define PAGE_NOT_PRESENT	0x5000
#define SST			0x6000 /* Single Step Trap */
#define SPECULATION		0x5700
#define TBT			0x5f00 /* Taken Branch Trap */
#define UNALIGNED_REF		0x5a00
#define UNSUPPORTED_DATA_REF	0x5b00
#define VHPT_TRANSLATION	0x0000

extern struct cpu {
	unsigned long psr;
	unsigned long cr[128];
	unsigned long kr[8];
	unsigned long rr[8];
	unsigned long gr[4]; /* r4-r7 */
	unsigned long saved_b2;
	unsigned long saved_b6;
	unsigned long saved_bank[16 * 2];
	unsigned int cfle;	/* Current frame load enable */
	unsigned int id;
	unsigned int irq_pending;
	unsigned int in_service_irq;
} cpu;

struct request {
	unsigned char type;
	unsigned char subtype;
	union {
		struct {
			unsigned long va;
			unsigned long itir;
		} purge;

		struct {
			unsigned long pa;
			int is_write;
		} invalidate;

		struct {
			unsigned long pa;
			int is_write;
		} fetch;

		struct {
			unsigned char irq;
		} ipi;

		struct {
			unsigned long ip;
			unsigned long gp;
		} startup;

		struct {
			unsigned long addr;
			unsigned long offset;
			unsigned int fd;
			unsigned int len;
		} ssc;
	};
};

#define IPC_FETCH		0
#define IPC_FETCH_REPLY		1
#define IPC_FETCH_NAK		2
#define IPC_INVALIDATE		3
#define IPC_INVALIDATE_REPLY	4
#define IPC_PURGE		5
#define IPC_PURGE_REPLY		6
#define IPC_IPI			7
#define IPC_STARTUP		8
#define IPC_SSC			9
#define IPC_SSC_REPLY		10
#define IPC_SHUTDOWN		11
#define SIGIPC 35
#define SIGSCSI (SIGIPC+1)
#define SIGKEYBD (SIGIPC+2)
#define SIGNETWK (SIGIPC+3)

struct siginfo;
struct sigcontext;


struct emulation_statistics {
	unsigned long	nbreak;
	unsigned long   ssm;
	unsigned long   rsm;
	unsigned long   mov_to_rr;
	unsigned long   mov_from_rr;
	unsigned long   mov_to_pkr;
	unsigned long   ptc_l;
	unsigned long   ptc_g;
	unsigned long   ptc_e;
	unsigned long   itr;
	unsigned long   nthash;
	unsigned long   nttag;
	unsigned long   mov_from_pmd;
	unsigned long   mov_from_pmc;
	unsigned long   ntpa;
	unsigned long   mov_from_cr;
	unsigned long   nread_psr;
	unsigned long   mov_to_ar;
	unsigned long   mov_to_cr;
	unsigned long   mov_to_psr;
	unsigned long   itc_d;
	unsigned long   itc_i;
	unsigned long   rfi;
	unsigned long   shrinkframes;
	unsigned long   bsw;

	unsigned long   redirects;
	unsigned long   faults;
	unsigned long   sigill;
	unsigned long   breaks;
	unsigned long   ticks;
	unsigned long   sigio;
	unsigned long   fpe;
	unsigned long   ssc;
	unsigned long   break_cover;
	unsigned long	break_dfh;
	unsigned long   break_halt;
	unsigned long	break_syscall;
	unsigned long   break_tpa;
	unsigned long   break_tlbop;
	unsigned long   break_rr;
	unsigned long   break_itm;
	unsigned long   break_rfi;
	unsigned long   break_irq;
	unsigned long   break_ivr;
	unsigned long   break_debug;
	unsigned long   break_pause;
};

extern struct emulation_statistics stats;

#ifdef EMULATOR_STAT
# define INC_STAT(stat) stats.stat++
#else
# define INC_STAT(stat) /* nothing */
#endif

extern unsigned long memsize; /* in bytes */
extern unsigned long pagesize; /* in bytes */

#define MEM_SIZE	memsize


extern int dumping;
extern jmp_buf continuation;

struct disk_stat {
	int fd;
	unsigned count;
};

struct disk_req {
	unsigned long addr;
	unsigned len;
};


extern int simscsi_irq;
extern pid_t mainpid;
extern int keyboard_irq, network_irq, network_fd;
extern int last_quiescent_req;
extern unsigned char network_ethaddr[6];
extern unsigned long cpu_freq;



#include "proto.h"
